################# Functions for AlphaVantage API #################

function get_price_data_single(currency::String)

    date = Dates.today()

    filename = "$(currency)_EUR_data_$(date).csv"
    filepath = joinpath("data", filename)

    df_raw = DataFrame()

    # Look for present day's CSV file, if not found, download and save data to a new file
    if isfile(filepath)
        @info "Reading $(currency) price/vol data from CSV file on disk"
        df_raw = CSV.File(filepath) |> DataFrame
    else
        try
            @info "Fetching $(currency) price/vol data from Alpha Vantage"
            raw = AlphaVantage.digital_currency_daily(currency, "EUR", datatype = "csv")
            df_raw = raw_to_df(raw)
            CSV.write(filepath, df_raw)
        catch
            @info "Could not fetch data, try again later!"
        end
    end

    # Return processed DataFrame only when raw data has been fetched successfully
    if ~isempty(df_raw)
        df_out_price, df_out_candle = average_df_price(currency, df_raw)
        df_out_vol = df_vol(currency, df_raw)

        return df_out_price, df_out_candle, df_out_vol
    else
        return DataFrame[], DataFrame[], DataFrame[]
    end
end

function get_ratings_data(currency::String)

    date = Dates.today()

    filename = "$(currency)_metrics_data_$(date).csv"
    filepath = joinpath("data", filename)

    df_metrics = DataFrame()

    # Look for present day's CSV file, if not found, download and save data to a new file
    if isfile(filepath)
        @info "Reading $(currency) FCAS data from CSV file on disk"
        df_metrics = CSV.File(filepath) |> DataFrame
    else
        try
            @info "Fetching $(currency) FCAS data from Alpha Vantage"
            rating = AlphaVantage.crypto_rating(currency)
            scores = rating["Crypto Rating (FCAS)"]
            CSV.write(filepath, scores)
            df_metrics = CSV.File(filepath) |> DataFrame
        catch err
            if isa(err, KeyError)
                @info "Could not retrieve data. Something wrong with the API, try again later!"
            else
                @info "This is a new error: $(err)"
            end
        end
    end

    ratings = ["utility", "fcas score", "developer", "market", "fcas rating"]
    index = Int64[]

    # Return scores only when metrics data has been fetched successfully
    if ~isempty(df_metrics)

        # Find the row(index) of a string match
        for rating in ratings
            i = findall(occursin.(rating, df_metrics[!, 1]))
            push!(index, i[1])
        end

        # Variables should be assigned in the same order as the list of ratings above
        utility_score, fcas_score, dev_score = df_metrics[!, 2][index[1]],
                                               df_metrics[!, 2][index[2]],
                                               df_metrics[!, 2][index[3]]

        mark_score, fcas_rating = df_metrics[!, 2][index[4]], 
                                  df_metrics[!, 2][index[5]]

        return utility_score, fcas_score, dev_score, mark_score, fcas_rating
    else
        return Number[], Number[], Number[], Number[], Number[]
    end

end